// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "network.hpp"

#include <netdb.h>
#include <sys/socket.h>

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <ostream>
#include <vector>

#include "Common.hpp"
#include "Debug.hpp"
#include "io.hpp"

#define MAX 1024

static void send_all(int fd, const void* buf, size_t len) {
	// emper send everything or encounters an unrecoverable state
	ssize_t bytes_send = emper::io::sendAndWait(fd, buf, len, MSG_NOSIGNAL, true);
	if (unlikely(bytes_send == -1)) {
		DIE_MSG_ERRNO("send failed");
	}
}

void echo_serve(int client_socket) {
	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	char buf[MAX];
	for (;;) {
		ssize_t bytes_recv = emper::io::recvAndWait(client_socket, buf, sizeof(buf), 0);
		// socket was shutdown
		if (bytes_recv == 0) {
			return;
		}

		if (bytes_recv == -1) {
			DIE_MSG_ERRNO("server read failed");
		}

		if (unlikely(strncmp("quit\n", buf, bytes_recv) == 0)) {
			exit(EXIT_SUCCESS);
		}

		send_all(client_socket, buf, bytes_recv);
	}
}

void echo_client(const std::string& host, const std::string& port,
								 const std::vector<std::string>& msgs, size_t iterations) {
	struct addrinfo* server;

	int err = getaddrinfo(host.c_str(), port.c_str(), nullptr, &server);
	if (err) {
		if (err == EAI_SYSTEM) {
			DIE_MSG_ERRNO("getaddrinfo failed");
		} else {
			LOGE("error in getaddrinfo: " << gai_strerror(err));
			exit(EXIT_FAILURE);
		}
	}

	int sockfd = socket(server->ai_family, server->ai_socktype, server->ai_protocol);
	if (sockfd == -1) {
		DIE_MSG_ERRNO("client socket creation failed");
	}

	if (emper::io::connectAndWait(sockfd, server->ai_addr, server->ai_addrlen) == -1) {
		DIE_MSG_ERRNO("connect failed");
	}

	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	char buf[MAX];
#ifndef NDEBUG
	memset(buf, 0, MAX);
#endif

	for (size_t i = 0; i < iterations; ++i) {
		for (const auto& s : msgs) {
			assert(s.length() <= MAX);
			send_all(sockfd, s.c_str(), s.length());

			ssize_t bytes_recv = emper::io::recvAndWait(sockfd, buf, s.length(), MSG_WAITALL);
			if (bytes_recv == 0) {
				DIE_MSG("client: socket closed");
			}

			if (bytes_recv == -1) {
				DIE_MSG_ERRNO("client read failed");
			}

			assert(bytes_recv > 0);
			assert((size_t)bytes_recv == s.length());

			if (memcmp(s.c_str(), buf, s.length()) != 0) {
				// if (strncmp(s.c_str(), buf, bytes_recv) != 0) {
				DIE_MSG("unexpected string received from server");
			}
		}
	}

	emper::io::shutdownAndWait(sockfd, SHUT_RDWR);
}
