// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include <sys/socket.h>

#include <array>
#include <chrono>
#include <compare>
#include <cstdlib>
#include <iostream>
#include <string>
#include <thread>
#include <vector>

#include "Common.hpp"
#include "CountingPrivateSemaphore.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "emper-config.h"
#include "emper.hpp"
#include "fixtures/network.hpp"
#include "io.hpp"

#define BUF_SIZE 1024

auto main(int argc, char* argv[]) -> int {
	if constexpr (!emper::IO || emper::IO_SYNCHRONOUS) {
		exit(77);
	}

	unsigned int client_count = std::thread::hardware_concurrency() * 4;
	unsigned int echos = EMPER_LOG_LEVEL > Info ? 250 : 25000;
	unsigned int computation = 0;
	std::string port = "4241";
	std::string host = "127.0.0.1";

	if (argc > 4) {
		std::cerr << "Usage: " << argv[0] << " [clients] [echos] [computation]" << std::endl;
		exit(EXIT_FAILURE);
	}

	const int DECIMAL = 10;
	if (argc > 1) {
		client_count = strtol(argv[1], nullptr, DECIMAL);
	}

	if (argc > 2) {
		echos = strtol(argv[2], nullptr, DECIMAL);
	}

	if (argc > 3) {
		computation = strtol(argv[3], nullptr, DECIMAL);
	}

	std::cout << "ConcurrentNetworkEchoTest with " << client_count << " clients sending " << echos
						<< " echos" << std::endl;

	Runtime runtime;

	auto serve_func = [&](int sock) {
		std::array<char, BUF_SIZE> buf;
		for (;;) {
			ssize_t bytes_recv = emper::io::recvAndWait(sock, buf.data(), buf.size(), 0);
			// socket was shutdown
			if (bytes_recv == 0) return;

			if (bytes_recv == -1) DIE_MSG_ERRNO("server read failed");

			if (computation) {
				const auto start = std::chrono::steady_clock::now();
				const auto deadline = start + std::chrono::microseconds(computation);
				// NOLINTNEXTLINE(modernize-use-nullptr)
				while (std::chrono::steady_clock::now() < deadline) {
				}
			}
			if (emper::io::sendAndWait(sock, buf.data(), bytes_recv, MSG_NOSIGNAL, true) == -1)
				DIE_MSG_ERRNO("send failed");
		}
	};

	Fiber* listener =
			emper::io::tcp_listener(host, port, serve_func, 1024, {emper::io::SockOpt::ReusePort});
	if (!listener) {
		DIE_MSG("failed to create listener");
	}

	runtime.scheduleFromAnywhere(*listener);

	Fiber* client_producer = Fiber::from([&] {
		CPS cps;
		for (unsigned int i = 0; i < client_count; ++i) {
			spawn(
					[&] {
						const std::vector<std::string> msgs = {"foo"};
						echo_client(host, port, msgs, echos);
					},
					cps);
		}
		cps.wait();
		runtime.initiateTermination();
	});

	runtime.scheduleFromAnywhere(*client_producer);

	runtime.waitUntilFinished();

	return EXIT_SUCCESS;
}
